//           ///          //                                        Mx clientSDK
//          /////        ////                    Mx Crossplatform Client Library
//         /// XXX     XXX ///
//        ///    XXX XXX    ///     $RCSfile: mxg2mjpg.cpp,v $
//       ///       XXX       ///     $Revision: 1.1 $
//      ///      XXX XXX      ///     $Date: 2006/01/17 12:43:19 $
//     ////    XXX     XXX    ////     $Author: cvs-kai $
//    ////                     ////
//   ////  M  O  B  O  T  I  X  ////////////////////////////////////////////////
//  //// Security Vision Systems //////////////////////////////////////////////
//                                                                          //
//  Copyright (C) 2005 - 2006, MOBOTIX AG, Germany                         //
//  This software is made available under the BSD licence. Please refer   //
//  to the file LICENCE.TXT contained in this distribution for details.  //
//                                                                      //
//  /////////////////////////////////////////////////////////////////////



/*! \file mxg2mjpg.cpp
    \brief The UNIX-style 
    <tt>MxPEG</tt>-to-<tt>MJPEG</tt>/<tt>RGB</tt>/<tt>yv12</tt> converter.
*/



#include <mxcpcFramewiseMxPEGDecoder.h>
#include <mxcpcStreamFile.h>
#include <mxcpcFramewiseMxPEGScanToJPEGConverterSoftwareOnly.h>
#include <mxcpcFramewiseMxPEGScanDecoderSoftwareOnly.h>
#include <mxcpcDefaultMxPEGDecoderBackEnd.h>
#include <mxcpcMxPEGRawFrameDumper.h>
#include <mxcpcMJPEGWriter.h>
#include <mxcpcDevNullStatusMessageHandler.h>
#include <mxcpcTime.h>

#include <cstdio>
#include <unistd.h>



void print_usage(const char *progname) {

  std::printf("%s : convert an MxPEG stream to MJPEG\n", progname);
  std::printf("%s\n", mxcpc::getVersionString());
  std::printf("\n");
  std::printf("Usage: %s [options]\n", progname);
  std::printf("\n");
  std::printf("Options:\n");
  std::printf("-h, --help             : this message\n");
  std::printf("-f <mxpeg_stream_file> : reads input from <mxpeg_stream_file> "
                                       "rather than from stdin\n");
  std::printf("-o <output_file>       : writes the generated MJPEG stream "
                                       "to <output_file> rather than to "
                                       "stdout\n");
  std::printf("-s <separator_string>  : HTTP mode, inserts a MIME separator "
                                       "between JPEG frames (useful in CGI "
                                       "scripts)\n");
  std::printf("-raw                   : raw mode, does not output MJPEG "
                                       "stream, but raw RGB data instead "
                                       "(3 bytes per pixel)\n");
  std::printf("-yuv                   : to be used with above raw mode, "
                                       "outputs frames in yv12 YUV format "
                                       "instead of RGB\n");
  std::printf("-sofps <framerate>     : to be used with above raw mode, "
                                       "stabilizes output framerate at the "
                                       "specified value\n");
  std::printf("\n");
  std::printf("Examples:\n");
  std::printf("cat file1.mxg file2.mxg | %s > file.mjpg\n", progname);
  std::printf("curl 'http://<camera_ip>/control/faststream.jpg?stream=MxPEG' "
              "| %s -s MOBOTIX_Fast_Serverpush -o file.mjpg\n", progname);
  std::printf("C:\\> type file1.mxg | %s > file.mjpg\n", progname);
  std::printf("\n"); 
}


void process_command_args(int argc, char **argv,
                          bool *show_usage,
                          const char **mime_separator,
                          const char **stream_filename,
                          const char **out_filename,
                          bool *raw,
                          bool *yuv,
                          float *sofs)                  {

  int arg_i;
  
  *show_usage      = false;
  *mime_separator  = 0;
  *stream_filename = 0;
  *out_filename    = 0;
  *raw             = false;
  *yuv             = false;
  *sofs            = -1.0f;
  
  arg_i = 1;
  while(arg_i < argc) {

    if(   !std::strcmp(argv[arg_i], "-h")
       || !std::strcmp(argv[arg_i], "--help")) {
      *show_usage = true;
      arg_i++;
    }
    else if(!std::strcmp(argv[arg_i], "-s")) {
      if(arg_i + 1 < argc) {
        *mime_separator = argv[arg_i + 1];
        arg_i += 2;
      }
      else {
        *show_usage = true;
      }
    }
    else if(!std::strcmp(argv[arg_i], "-f")) {
      if(arg_i + 1 < argc) {
        *stream_filename = argv[arg_i + 1];
        arg_i += 2;
      }
      else {
        *show_usage = true;
      }
    }
    else if(!std::strcmp(argv[arg_i], "-o")) {
      if(arg_i + 1 < argc) {
        *out_filename = argv[arg_i + 1];
        arg_i += 2;
      }
      else {
        *show_usage = true;
      }
    }
    else if(!std::strcmp(argv[arg_i], "-raw")) {
      *raw = true;
      arg_i ++;
    }
    else if(!std::strcmp(argv[arg_i], "-yuv")) {
      *yuv = true;
      arg_i ++;
    }
    else if(!std::strcmp(argv[arg_i], "-sofps")) {
      if(arg_i + 1 < argc) {
        *sofs = std::atof(argv[arg_i + 1]);
        if(*sofs < .01f)    *sofs =    .01f;
        if(*sofs > 1000.0f) *sofs = 1000.0f;
        arg_i += 2;
      }
      else {
        *show_usage = true;
      }
    }
    else {
      *show_usage = true;
    }
    
    if(*show_usage) break;
  }
}



int main(int argc, char **argv) {

  bool show_usage;
  const char *mime_separator;
  const char *stream_filename;
  const char *out_filename;
  const char *progname;
  bool raw;
  bool yuv;
  float sofs;
  mxcpcFramewiseMxPEGDecoder *decoder;
  mxcpcStreamFile *stream_file;
  mxcpcMJPEGWriter *writer;
  mxcpcMxPEGRawFrameDumper *frame_dumper;
  mxcpcBinaryFileWriter *raw_writer;
  mxcpcTime da_time;
  int cycle_time, time_left, timeout;
  
  progname = "mxg2mjpg";
  
  // process command args...
  process_command_args(argc, argv,
                       &show_usage, 
                       &mime_separator, 
                       &stream_filename, 
                       &out_filename,
                       &raw,
                       &yuv,
                       &sofs);
  if(show_usage) {
    print_usage(progname);
    std::exit(0);
  }

  // ok, we actually have some work to do...
  
  // we don't have use for mxcpc status messages...
  mxcpc::setStatusMessageHandler(new mxcpcDevNullStatusMessageHandler());
  
  // open input streamfile...
  if(stream_filename) 
   stream_file = new mxcpcStreamFile(stream_filename);
  else
   stream_file = new mxcpcStreamFile(STDIN_FILENO);
  
  // open output...
  writer     = 0;   // suppress compiler warnings...
  raw_writer = 0;
  if(!raw) {
    if(out_filename)
     writer = new mxcpcMJPEGWriter(out_filename, mime_separator);
    else
     writer = new mxcpcMJPEGWriter(STDOUT_FILENO, mime_separator);
  }
  else {
    if(out_filename)
     raw_writer = new mxcpcBinaryFileWriter(out_filename);
    else
     raw_writer = new mxcpcBinaryFileWriter(STDOUT_FILENO);
  }
  
  // set up decoder configuration...
  if(!raw) {
    decoder = new mxcpcFramewiseMxPEGDecoder(
                    stream_file,
                    new mxcpcDefaultMxPEGDecoderBackEnd(),
                    new mxcpcFramewiseMxPEGScanToJPEGConverterSoftwareOnly(
                         writer
                        )
                  );
  }
  else {
    frame_dumper = new mxcpcMxPEGRawFrameDumper(raw_writer, yuv);
    decoder = new mxcpcFramewiseMxPEGDecoder(
                    stream_file,
                    frame_dumper,
                    new mxcpcFramewiseMxPEGScanDecoderSoftwareOnly()
                  );
  }
  
  // process stream data...
  if(raw & sofs > 0.0f) {
  
    cycle_time = (int)(1000.0f / sofs);
    time_left  = 0;
    
    frame_dumper->setExplicitDumpModeEnabled(true);
    
    while(decoder->sourceStillUp()) {
      
      // check time...
      time_left -= da_time.elapsed();
      
      // dump due frames...
      while(time_left <= 0) {
        frame_dumper->dumpFrame();
        time_left += cycle_time;
      }
      
      // determine timeout so we get wakened in time for the next frame to 
      // dump...
      timeout = time_left;
      if(timeout < 10)   timeout = 10;
      if(timeout > 1000) timeout = 1000;   // don't block longer than a second
      stream_file->setTimedBlockModeEnabled(true, timeout);
      
      // decode or block until timeout...
      decoder->decode(2048);
    }
  
  }
  else {
    while(decoder->sourceStillUp()) 
     decoder->decode(2048); 
  }
  
  // close files, clean up and and exit...
  delete decoder;
  mxcpc::setStatusMessageHandler(0);
  std::exit(0);
}

